from xasp.entities import Explain
from dumbo_asp.primitives.models import Model

explain = Explain.the_program(
    "a :- not b. b :- not a.",
    the_answer_set=Model.of_atoms("a"),
    the_atoms_to_explain=Model.of_atoms("a"),
)
explain.compute_explanation_dag()  # it can also be omitted
print(explain.explanation_dag())
explain.show_navigator_graph()